; Z88 CamelForth
; cameldep.s (model-dependent words)

; Note that although the original Z80 CamelForth released by Bradford J.
; Rodriguez was fairly easily portable to other machines and even processors,
; this is no longer really the case due to the massive amount of extra
; Z80 & Z88-specific words and optimisations added.

;C ALIGN    --                         align HERE
    head ALIGN,5,"ALIGN",docolon,1,0
	defw EXIT

;C ALIGNED  addr -- a-addr       align given addr
    head ALIGNED,7,"ALIGNED",docolon,1,0
	defw EXIT

;C CELL+    a-addr1 -- a-addr2      add cell size
;   2 + ;
    head CELLPLUS,5,"CELL+",docode,0,celpln-celplf
.celplf inc bc
        inc bc
.celpln
	next

;C CELLS    n1 -- n2            cells->adrs units
    head CELLS,5,"CELLS",docode,0,cellsn-cellsf
.cellsf sla c
	rl b
.cellsn
	next

;C CHAR+    c-addr1 -- c-addr2   add char size
    head CHARPLUS,5,"CHAR+",docode,0,chplsn-chplsf
.chplsf inc bc
.chplsn
	next

;C CHARS    n1 -- n2            chars->adrs units
    head CHARS,5,"CHARS",docolon,1,0
	defw EXIT

;C >BODY    xt -- a-addr      adrs of param field
;   3 + @ ;                     Z80 (3 byte CALL)
    head TOBODY,5,">BODY",docode,0,tobodn-tobodf
.tobodf ld hl,3
	add hl,bc
	ld c,(hl)
	inc hl
	ld b,(hl)
.tobodn
	next


; I called this word ,XT before I discovered that
; it is defined in the ANSI standard as COMPILE,.
; On a DTC Forth this simply appends xt (like , )
; but on an STC Forth this must append 'CALL xt'.
; For Z88 CamelForth, in standard compile mode we just compile the xt,
; but in "fast" mode we copy the required code or cause an exception.

;X COMPILE,  xt --         append execution token
;   usr_fastcomp C@
;   IF  DUP CFA>FCA C@ 127 AND
;       ?DUP 0= IF  DUP C@ CD =
;		    OVER @ DUP docon =
;                   SWAP docreate = OR OVER AND
;		    IF  DROP CFA>PFA @ ,LIT
;                   ELSE  OVER @ dovalue = AND
;			  IF  CFA>PFA @
;			      C5 C, 4BED , ,
;			  ELSE  -4095 THROW
;			  THEN
;                   THEN
;               ELSE
;                   HERE OVER ALLOT SWAP CMOVE
;               THEN
;   ELSE  ,
;   THEN ;
head COMMAXT,8,"COMPILE"&$2c,docode,0,0
.COMMABRANCH
	ld a,(usr_fastcomp)
	and a
	jp z,COMMA
	push bc
        dec bc
        ld a,(bc)
        and 127
	jr z,CMXT2
	exx
	ld b,0
	ld c,a				; BC=codesize
	ld hl,(usr_curds)
	ld e,(hl)
	inc hl
	ld d,(hl)			; DE=DP
	ex (sp),hl			; HL=code address, TOS=add of DP+1
	ldir				; do the copy
	pop hl
	ld (hl),d
	dec hl
	ld (hl),e			; set new DP
	exx
	pop bc				; get new TOS
	next
.CMXT2	pop hl
	ld a,(hl)
	cp $cd
	jr nz,CMXT3
	inc hl
	ld c,(hl)
	inc hl
	ld b,(hl)		; BC=code address
	inc hl
	push hl
	ld hl,docon
	and a
	sbc hl,bc
	jr z,CMXT4
	ld hl,docreate
	and a
	sbc hl,bc
	jr z,CMXT4
	ld hl,dovalue
	and a
	sbc hl,bc
	jr nz,CMXT3
	pop hl
	ld c,(hl)
	inc hl
	ld b,(hl)		; BC=value address
	push bc
	ld bc,$4bed		; ld bc,()
	push bc
	ld bc,$c5		; push bc
	call docolon
	defw CCOMMA,COMMA,COMMA,EXIT
.CMXT4	pop hl
	ld c,(hl)
	inc hl
	ld b,(hl)		; BC=constant value/variable address
	jp COMMALIT
.CMXT3	ld bc,-4095
	jp THROW		; error if "unfastable" word

;Z !CF    adrs cfa --   set code action of a word
;   0CD OVER C!         store 'CALL adrs' instr
;   1+ ! ;              Z80 VERSION
; Depending on the implementation this could
; append CALL adrs or JUMP adrs.
    nohead STORECF,docode
	ld h,b
	ld l,c
	pop bc
	ld (hl),$0cd
	inc hl
	ld (hl),c
	inc hl
	ld (hl),b
	pop bc
	next

;Z ,CF    adrs --       append a code field
;   HERE !CF 3 ALLOT ;  Z80 VERSION (3 bytes)
    nohead COMMACF,docolon
        defw HERE,STORECF,LIT,3,ALLOT,EXIT

;Z !COLON   --      change code field to docolon
;   docolon-adrs LATEST @ NFA>CFA !CF -1 CELLS ALLOT ;
; This should be used immediately after CREATE.
; Colon definitions don't need a pointer to the body, as
; codespace is always the same as namespace
; This is made a distinct word, because on an STC
; Forth, colon definitions have no code field.
    nohead STORCOLON,docolon
        defw LIT,docolon,LATEST,FETCH,NFATOCFA
	defw STORECF,LIT,-2,ALLOT,EXIT

;C IMMEDIATE	--		make last definition immediate
    head IMMEDIATE,9,"IMMEDIATE",docode,0,0
	ld hl,(usr_latest)
	ld a,(hl)
	inc a
	add a,l
	ld l,a
	ld a,h
	adc a,0
	ld h,a
	set 7,(hl)
	next

; CONTROL STRUCTURES ============================
; These words allow Forth control structure words
; to be defined portably.

;Z ,BRANCH   xt --    append a branch instruction
; xt is the branch operator to use, e.g. qbranch
; or (loop).  It does NOT append the destination
; address.  On the Z80 this is equivalent to ,XT.

;Z ,DEST   dest --        append a branch address
; This appends the given destination address to
; the branch instruction.  On the Z80 this is ','
; ...other CPUs may use relative addressing.
; Not defined for Z88 - just equates to COMMA (in camelhi.s)


;Z !DEST   dest adrs --    change a branch dest'n
; Changes the destination address found at 'adrs'
; to the given 'dest'.  On the Z80 this is '!'
; ...other CPUs may need relative addressing.


; MACHINE CODE TOOLS ============================

;X CODE		"<spaces>name" --	Create code definition
;   NSDS CREATE -5 ALLOT ;
    head CODE,4,"CODE",docolon,0,0
	defw NSDS,CREATE,LIT,-5,ALLOT,EXIT

;Z NEXT		--		Finish code definition
;   0C3 C, next_code , ;
    head NEXT,4,"NEXT",docolon,0,0
	defw LIT,$0C3,CCOMMA
	defw LIT,next_code,COMMA,EXIT

;Z FCODE	"<spaces>name" -- cfa		Create "fast" code definition
;   CODE HERE ;
    head FCODE,5,"FCODE",docolon,0,0
        defw CODE,HERE,EXIT

;Z FNEXT	cfa --				Finish "fast" code definition
;   HERE OVER - SWAP 1 CHARS - C! NEXT ;
    head FNEXT,5,"FNEXT",docolon,0,0
        defw HERE,OVER,MINUS,SWOP,ONEMINUS,CSTORE,NEXT,EXIT

;Z (ENDFAST)	--				Exit fast inline code and
;						continue with colon definition
; This is a special definition which is executed with a call from inline code
.BRENDFAST
	pop de				; IP=return address
	innext				; start executing again

;Z {		--				Start inlining fast code
;    1 usr_fastcomp C!
;    HERE CELL+ , ; IMMEDIATE
    head LCURLY,1,"{",docolon,1,0
	defw ONE,LIT,usr_fastcomp,CSTORE
	defw HERE,CELLPLUS,COMMA,EXIT

;Z }		--				End inlining fast code
;    0 usr_fastcomp C!
;    CD C, ['] (ENDFAST) , ; IMMEDIATE
    head RCURLY,1,"}",docolon,1,0
	defw ZERO,LIT,usr_fastcomp,CSTORE
	defw LIT,$cd,CCOMMA,LIT,BRENDFAST,COMMA,EXIT

