; Z88 CamelForth
; camelhi.s (mostly high-level system-independent words)

; SYSTEM VARIABLES & CONSTANTS ==================

;C BL      -- char            an ASCII space
    head BL,2,"BL",docon,0,0
        defw $20

;X tib     -- a-addr     Terminal Input Buffer
    head TIB,3,"TIB",docon,0,0
        defw usr_tib

;X TRUE    -- true	True flag
;  -1 CONSTANT TRUE
    head TRUE,4,"TRUE",docon,0,0
	defw -1

;X FALSE   -- false	False flag
;   0 CONSTANT FALSE
    head FALSE,5,"FALSE",docon,0,0
	defw 0

;Z U0      -- a-addr       current user area adrs
;  0 USER U0
    head U0,2,"U0",docon,0,0
        defw userst

;C >IN     -- a-addr        holds offset into TIB
    head TOIN,3,">IN",docon,0,0
        defw usr_in

;C BASE    -- a-addr       holds conversion radix
    head BASE,4,"BASE",docode,0,basen-basef
.basef	push bc
	push iy
	pop hl
	ld bc,usr_base
	add hl,bc
	ld b,h
	ld c,l
.basen
	next

;C STATE   -- a-addr       holds compiler state
;  This holds 0 in interpret mode and -1 in compile mode
;  In compile mode, the usr_fastcomp variable determines whether fast compilation
;  is on.
    head STATE,5,"STATE",docon,0,0
    	defw usr_state

;Z FAST?   -- flag	  return fast code compilation mode (0 or 1)
   head FASTQ,5,"FAST?",docode,0,0
	push bc
	ld a,(usr_fastcomp)
	ld c,a
	ld b,0
	next

;Z INACC?   -- c-addr       byte variable, holds one if ACCEPT executing
   head INACCEPT,6,"INACC?",docon,0,0
	defw usr_inaccept

;Z CURNS    -- a-addr	    holds current name/codespace pointer
   nohead CURNS,docon
	defw usr_curns

;Z CURDS    -- a-addr	    holds current dataspace pointer
   nohead CURDS,docon
	defw usr_curds

;Z DP	    -- a-addr	    address of current dataspace pointer
;  CURDS @ ;
   head DP,2,"DP",docode,0,dpn-dpf
.dpf    push bc
	ld bc,(usr_curds)
.dpn
	next


;Z 'SOURCE  -- a-addr      two cells: len, adrs
     nohead TICKSOURCE,docon
	defw usr_source

;Z LATEST    -- a-addr     last word in dict.
    head LATEST,6,"LATEST",docon,0,0
        defw usr_latest

;Z LP       -- a-addr     Leave-stack pointer
    nohead LP,docode
	push bc
	push iy
	pop hl
	ld bc,usr_lp
	add hl,bc
	ld b,h
	ld c,l
	next

;Z 'SOURCE-ID  -- a-addr
    nohead TICKSOURCEID,docon
    	defw usr_srcid

;X BLK		-- a-addr
    head BLK,3,"BLK",docon,0,0
	defw usr_blk

;X SOURCE-ID    -- sourceid
;   'SOURCE-ID @ ;
    head SOURCEID,9,"SOURCE-ID",docode,0,srcidn-srcidf
.srcidf push bc
	ld bc,(usr_srcid)
.srcidn
	next

;X PAD       -- a-addr    user PAD buffer
;                         = end of hold area!
    head PAD,3,"PAD",docon,0,0
        defw usr_pad


; ARITHMETIC OPERATORS ==========================

;C S>D    n -- d          single -> double prec.
;   DUP 0< ;
    head STOD,3,"S>D",docolon,0,0
        defw DUP,ZEROLESS,EXIT

;Z ?NEGATE  n1 n2 -- n3  negate n1 if n2 negative
;   0< IF NEGATE THEN ;        ...a common factor
    nohead QNEGATE,docode
	bit 7,b
	pop bc
	jr z,qnegn
	ld a,b
	cpl
	ld b,a
	ld a,c
	cpl
	ld c,a
	inc bc
.qnegn
	next

;C ABS     n1 -- +n2     absolute value
;   DUP ?NEGATE ;
    head ABS,3,"ABS",docolon,0,0
        defw DUP,QNEGATE,EXIT

;X DNEGATE   d1 -- d2     negate double precision
;   SWAP INVERT SWAP INVERT 1 M+ ;
    head DNEGATE,7,"DNEGATE",docolon,0,0
        defw SWOP,INVERT,SWOP,INVERT,ONE,MPLUS
        defw EXIT

;Z ?DNEGATE  d1 n -- d2   negate d1 if n negative
;   0< IF DNEGATE THEN ;       ...a common factor
    nohead QDNEGATE,docolon
        defw ZEROLESS,QBRANCH,DNEG1,DNEGATE
.DNEG1  defw EXIT

;X DABS     d1 -- +d2    absolute value dbl.prec.
;   DUP ?DNEGATE ;
    head DABS,4,"DABS",docolon,0,0
        defw DUP,QDNEGATE,EXIT

;C M*     n1 n2 -- d    signed 16*16->32 multiply
;   2DUP XOR >R        carries sign of the result
;   SWAP ABS SWAP ABS UM*
;   R> ?DNEGATE ;
    head MSTAR,2,"M*",docolon,0,0
        defw TWODUP,XOR,TOR
        defw SWOP,ABS,SWOP,ABS,UMSTAR
        defw RFROM,QDNEGATE,EXIT

;C SM/REM   d1 n1 -- n2 n3   symmetric signed div
;   2DUP XOR >R              sign of quotient
;   OVER >R                  sign of remainder
;   ABS >R DABS R> UM/MOD
;   SWAP R> ?NEGATE
;   SWAP R> ?NEGATE ;
; Ref. dpANS-6 section 3.2.2.1.
    head SMSLASHREM,6,"SM/REM",docolon,0,0
        defw TWODUP,XOR,TOR,OVER,TOR
        defw ABS,TOR,DABS,RFROM,UMSLASHMOD
        defw SWOP,RFROM,QNEGATE,SWOP,RFROM,QNEGATE
        defw EXIT

;C FM/MOD   d1 n1 -- n2 n3   floored signed div'n
;   DUP >R              save divisor
;   SM/REM
;   OVER 0<> OVER 0< AND
;   IF           if quotient negative AND remainder not zero
;       SWAP R> +         add divisor to rem'dr
;       SWAP 1-           decrement quotient
;   ELSE R> DROP THEN ;
; Ref. dpANS-6 section 3.2.2.1.
    head FMSLASHMOD,6,"FM/MOD",docolon,0,0
        defw DUP,TOR,SMSLASHREM
        defw OVER,ZERONOTEQUAL,OVER,ZEROLESS,AND,QBRANCH,FMMOD1
        defw SWOP,RFROM,PLUS,SWOP,ONEMINUS
        defw EXIT
.FMMOD1 defw RFROM,DROP
.FMMOD2 defw EXIT

;C *      n1 n2 -- n3       signed multiply
;   M* DROP ;
    head STAR,1,"*",docolon,0,0
        defw MSTAR,DROP,EXIT

;C /MOD   n1 n2 -- n3 n4    signed divide/rem'dr
;   >R S>D R> FM/MOD ;
    head SLASHMOD,4,"/MOD",docolon,0,0
        defw TOR,STOD,RFROM,FMSLASHMOD,EXIT

;C /      n1 n2 -- n3       signed divide
;   /MOD nip ;
    head SLASH,1,"/",docolon,0,0
        defw SLASHMOD,NIP,EXIT

;C MOD    n1 n2 -- n3       signed remainder
;   /MOD DROP ;
    head MOD,3,"MOD",docolon,0,0
        defw SLASHMOD,DROP,EXIT

;C */MOD  n1 n2 n3 -- n4 n5    n1*n2/n3, rem&quot
;   >R M* R> FM/MOD ;
    head SSMOD,5,"*/MOD",docolon,0,0
        defw TOR,MSTAR,RFROM,FMSLASHMOD,EXIT

;C */     n1 n2 n3 -- n4        n1*n2/n3
;   */MOD nip ;
    head STARSLASH,2,"*/",docolon,0,0
        defw SSMOD,NIP,EXIT

;C MAX    n1 n2 -- n3       signed maximum
;   2DUP < IF SWAP THEN DROP ;
    head MAX,3,"MAX",docolon,0,0
        defw TWODUP,LESS,QBRANCH,MAX1,SWOP
.MAX1   defw DROP,EXIT

;C MIN    n1 n2 -- n3       signed minimum
;   2DUP > IF SWAP THEN DROP ;
    head MIN,3,"MIN",docolon,0,0
        defw TWODUP,GREATER,QBRANCH,MIN1,SWOP
.MIN1   defw DROP,EXIT


; TRIGONOMETRIC FUNCTIONS

;Z sines	-- a-addr	table of sines (hidden)
;   CREATE sines .. , .... ;
    nohead sines,docreate
	defw 0000,0175,0349,0523,0698,0872,1045
	defw 1219,1392,1564,1736,1908,2079,2250
	defw 2419,2588,2756,2924,3090,3256,3420
	defw 3548,3746,3907,4067,4226,4384,4540
	defw 4695,4848,5000,5150,5229,5446,5592
	defw 5736,5878,6018,6157,6293,6428,6561
	defw 6691,6820,6947,7071,7193,7313,7431
	defw 7547,7660,7771,7880,7986,8090,8191
	defw 8290,8387,8480,8572,8660,8746,8829
	defw 8910,8988,9063,9135,9205,9272,9336
	defw 9397,9455,9511,9563,9613,9659,9703
	defw 9744,9781,9816,9848,9877,9903,9926
	defw 9956,9962,9976,9986,9994,9998,10000

;Z SIN		angle -- sin*10000
;   ABS 90 /MOD
;   TUCK 1 AND IF  90 SWAP -  THEN
;   CELLS sines + @
;   SWAP 2 AND IF  NEGATE  THEN ;
    head SIN,3,"SIN",docolon,0,0
	defw ABS,LIT,90,SLASHMOD
	defw TUCK,ONE,AND,QBRANCH,sin1
	defw LIT,90,SWOP,MINUS
.sin1   defw CELLS,sines,PLUS,FETCH
	defw SWOP,LIT,2,AND,QBRANCH,sin2
	defw NEGATE
.sin2   defw EXIT

;Z COS		angle -- cos*10000
;   90 + SIN ;
    head COS,3,"COS",docolon,0,0
	defw LIT,90,PLUS,SIN,EXIT

;Z SIN*		n angle -- n*sin(angle)
;   SIN 10000 */ ;
    head SINSTAR,4,"SIN*",docolon,0,0
	defw SIN,LIT,10000,STARSLASH,EXIT

;Z COS*		n angle -- n*cos(angle)
;   COS 10000 */ ;
    head COSSTAR,4,"COS*",docolon,0,0
	defw COS,LIT,10000,STARSLASH,EXIT


; DOUBLE OPERATORS ==============================

;C 2@    a-addr -- x1 x2    fetch 2 cells
;   DUP CELL+ @ SWAP @ ;
;   the lower address will appear on top of stack
    head TWOFETCH,2,"2@",docode,0,twoftn-twoftf
.twoftf	ld h,b
	ld l,c
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	push hl
.twoftn
	next

;C 2!    x1 x2 a-addr --    store 2 cells
;   SWAP OVER ! CELL+ ! ;
;   the top of stack is stored at the lower adrs
    head TWOSTORE,2,"2!",docode,0,twostn-twostf
.twostf	ld h,b
	ld l,c
	pop bc
	ld (hl),c
	inc hl
	ld (hl),b
	inc hl
	pop bc
	ld (hl),c
	inc hl
	ld (hl),b
	pop bc
.twostn
	next

;C 2DROP  x1 x2 --          drop 2 cells
;   DROP DROP ;
    head TWODROP,5,"2DROP",docode,0,tdropn-tdropf
.tdropf pop bc
	pop bc
.tdropn
	next

;C 2DUP   x1 x2 -- x1 x2 x1 x2   dup top 2 cells
;   OVER OVER ;
    head TWODUP,4,"2DUP",docode,0,tdupn-tdupf
.tdupf  pop hl
	push hl
	push bc
	push hl
.tdupn
	next

;C 2SWAP  x1 x2 x3 x4 -- x3 x4 x1 x2  per diagram
;   ROT >R ROT R> ;
    head TWOSWAP,5,"2SWAP",docode,0,tswapn-tswapf
.tswapf pop hl  ; bc=x4, hl=x3
	exx
	pop bc
	pop hl	; bc'=x2, hl'=x1
	exx
	push hl
	push bc
	exx
	push hl
	push bc
	exx
	pop bc
.tswapn
	next

;C 2OVER  x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2
;   >R >R 2DUP R> R> 2SWAP ;
    head TWOOVER,5,"2OVER",docode,0,tovern-toverf
.toverf	exx
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl
	exx
	push bc
	exx
	push bc
	push de
	exx
	pop bc
.tovern
	next

; INPUT/OUTPUT ==================================

;C COUNT   c-addr1 -- c-addr2 u  counted->adr/len
;   DUP CHAR+ SWAP C@ ;
    head COUNT,5,"COUNT",docode,0,counn-counf
.counf  ld a,(bc)
	inc bc
	push bc
	ld c,a
	ld b,0
.counn
	next

;C CR      --               output newline
;   0D EMIT 0A EMIT ;
    head CRR,2,"CR",docolon,0,0
        defw LIT,$0d,EMIT,LIT,$0a,EMIT,EXIT

;C SPACE   --               output a space
;   BL EMIT ;
    head SPACE,5,"SPACE",docolon,0,0
        defw BL,EMIT,EXIT

;C SPACES   n --            output n spaces
;   BEGIN DUP WHILE SPACE 1- REPEAT DROP ;
    head SPACES,6,"SPACES",docolon,0,0
.SPCS1  defw DUP,QBRANCH,SPCS2
        defw SPACE,ONEMINUS,BRANCH,SPCS1
.SPCS2  defw DROP,EXIT

;Z UMIN     u1 u2 -- u      unsigned minimum
;   2DUP U> IF SWAP THEN DROP ;
    head UMIN,4,"UMIN",docolon,0,0
        defw TWODUP,UGREATER,QBRANCH,UMIN1,SWOP
.UMIN1  defw DROP,EXIT

;Z UMAX    u1 u2 -- u       unsigned maximum
;   2DUP U< IF SWAP THEN DROP ;
    head UMAX,4,"UMAX",docolon,0,0
        defw TWODUP,ULESS,QBRANCH,UMAX1,SWOP
.UMAX1  defw DROP,EXIT


;Z ACCEPT	caddr +n|-n -- +n'	Get line from terminal
;   INACC? C@ ROT ROT 1 INACC? C!	If n<0, uses current contents of buffer
;   DUP 0< IF  NEGATE 2DUP -TRAILING TUCK TYPE
;          ELSE  0  THEN  ROT OVER	(S: maxlen curlen addr pos )
;   (ACC_MAIL)
;   PENDING-CHAR @ DUP 0<  -1 PENDING-CHAR !
; if DROP
;   BEGIN  EKEY
; then
;          DUP 13 <>
;   WHILE  DUP 31 U> OVER 127 U< AND
;          IF  >R 2OVER U>
;              IF  R@ CEMIT
;                  DUP 2OVER ROT TUCK + DUP 1+ 2SWAP - CMOVE>
;                  2DUP + R> SWAP C!
;                  1+ ROT 1+
;                  2DUP SWAP - ?DUP
;                  IF  DUP >R 2OVER + R@ TYPE R> 8 MEMIT  THEN
;                  ROT ROT
;              ELSE  R> DROP  THEN
;          ELSE  2DUP 8 = SWAP 0> AND
;                IF  DROP DUP 2OVER ROT TUCK + DUP 1- 2SWAP - CMOVE
;                    1- ROT 1-
;                    2DUP SWAP - ?DUP
;                    IF  DUP >R 2OVER + R@ TYPE R> 8 MEMIT  THEN
;                    ROT ROT
;                ELSE  DUP 508 =
;                IF  DROP DUP IF  1- 8 CEMIT  THEN  
;                ELSE  DUP 509 =
;		 IF  DROP 2OVER NIP OVER U> IF  1+ 9 CEMIT  THEN
;		 ELSE  DUP 500 =
;		 IF  DROP 8 MEMIT 0
;		 ELSE  DUP 501 =
;		 IF  DROP >R OVER DUP R> - 9 MEMIT
;		 ELSE  DUP 256 =
;                IF  DROP (ACC_MAIL)
;                ELSE  0 INACCEPT C! (ACC_EVT) 1 INACCEPT C!
;                THEN THEN THEN THEN THEN
;          THEN
;   REPEAT  SWAP INACCEPT C!
;   2DROP DROP NIP DUP R> - 9 MEMIT ; 	; move cursor to end of input
    head ACCEPT,6,"ACCEPT",docolon,0,0
	defw INACCEPT,CFETCH,ROT,ROT
	defw DUP,ZEROLESS,QBRANCH,cacc1
	defw NEGATE,TWODUP,MINUSTRAILING,TUCK,TYPE,BRANCH,cacc2
.cacc1	defw ZERO
.cacc2	defw ROT,OVER,BRACCMAIL
	defw PENDINGCHAR,FETCH,DUP,ZEROLESS,MINUSONE,PENDINGCHAR,STORE
	defw QBRANCH,caccA,DROP
.cacc335 defw ONE,INACCEPT,CSTORE
.cacc34	defw EKEY
.caccA	defw caccA2
.caccA2	ld a,b
	and a
	ld a,c
	jr nz,cacot2
	cp 13
	jp z,cacend		; out of loop if CR
	cp 32
	jr c,cacoth
	cp 128
	jr nc,cacoth		; by this stage, 32<=char=<127
	pop bc
	pop hl
	exx
	pop bc
	pop hl
	push hl
	push bc
	and a
	sbc hl,bc
	exx
	push hl
	jr z,cacclp		; do nothing if maxlen=curlen
	pop hl
	pop de
	inc de
	push de
	push hl
	inc bc
	push bc			; S: maxlen curlen+1 addr pos+1
	add hl,bc
	dec hl			; HL=addr+pos
	ex de,hl
	and a
	sbc hl,bc
	ex de,hl		; DE=curlen-pos
	push de
.inschr	ld c,(hl)		; get next char to move up
	ld (hl),a		; insert a character and emit it
	call_oz(os_out)
	inc hl
	ld a,d
	or e
	jr z,endins
	dec de
	ld a,c
	jr inschr
.endins	pop hl
.enddel	ld c,8
	call memit		; move cursor back to required position
.cacpop	pop bc			; BC=TOS again
.cacclp	ld de,caccA
	jp EKEY			; next key and loop back
.cacoth	cp 8
	jp nz,cacevt		; only 8 dealt with if B=0
	pop bc
	ld a,b
	or c
	jr z,cacclp		; do nothing if pos=0
	pop hl
	pop de
	dec de
	push de
	push hl
	dec bc
	push bc			; S: maxlen curlen-1 addr pos-1
	add hl,bc
	inc hl			; HL=addr+pos
	ex de,hl
	and a
	sbc hl,bc
	ex de,hl		; DE=curlen-pos
	push de
	ld a,8			; go back 1 pos
.dellp	call_oz(os_out)		; emit char
	ld a,d
	or e
	jr z,findel
	dec de
	ld a,(hl)
	dec hl
	ld (hl),a		; move char back
	inc hl
	inc hl
	jr dellp
.findel	ld a,32
	call_oz(os_out)		; emit a space to erase final character
	pop hl
	inc hl
	jr enddel
.cacot2	cp 508-256
	jr z,cac508
	cp 509-256
	jr z,cac509
	cp 500-256
	jr z,cac500
	cp 501-256
	jr z,cac501
	and a
	jr nz,cacevt
.cac256 pop bc
	ld de,cacc34
	jp BRACCMAIL		; DROP (ACC_MAIL)
.cac501	pop bc			; BC=pos
	pop de			; DE=addr
	pop hl			; HL=curlen
	push hl
	push de
	push hl			; S: maxlen curlen addr curlen
	and a
	sbc hl,bc
	ld c,9
	call memit		; move cursor right
	jr cacpop
.cac500	pop hl
	ld c,8
	call memit		; move cursor left
	ld bc,0
	jr cacclp
.cac508 pop bc
	ld a,b
	or c
	jr z,cacclp
	dec bc
	ld a,8
	call_oz(os_out)
	jr cacclp
.cac509	pop bc			; BC=pos
	pop de			; DE=addr
	pop hl			; HL=curlen
	push hl
	push de
	and a
	sbc hl,bc
	jr c,dcaclp
	jr z,dcaclp
	inc bc
	ld a,9
	call_oz(os_out)
.dcaclp	jp cacclp
.cacevt	xor a
	ld (usr_inaccept),a	
	ld de,cacc335
	jp BRACCEVT		; (ACC_EVT) and loop back
.cacend	pop bc			; BC=pos
	pop af
	pop hl			; HL=curlen
	pop af
	push hl
	and a
	sbc hl,bc		; HL=curlen-pos
	ld c,9
	call memit		; get to end of line
	pop bc			; BC=curlen
	pop hl			; L=original INACCEPT
	ld a,l
	ld (usr_inaccept),a	; reset INACCEPT
	jp EXIT


; Subroutine to emit HL repetitions of char in C

.memit	ld a,h
	or l
	ret z
	ld a,c
	call_oz(os_out)
	dec hl
	jr memit

;Z (S")     -- c-addr u   run-time code for S"
;   R> COUNT 2DUP + ALIGNED >R  ;
    nohead XSQUOTE,docolon
        defw RFROM,COUNT,TWODUP,PLUS,TOR
        defw EXIT

;Z (")		"ccc<quote>" --
;   [CHAR] " PARSE
;   DUP C, HERE SWAP DUP ALLOT CMOVE ;
    nohead BRQUOTE,docolon
	defw LIT,$22,PARSE
	defw DUP,CCOMMA,HERE,SWOP,DUP,ALLOT,CMOVE,EXIT

;C ."       --         compile string to print
;   POSTPONE S"  POSTPONE TYPE ; IMMEDIATE
    head DOTQUOTE,2,"."&$22,docolon,1,0
        defw SQUOTE
        defw LIT,TYPE,COMMAXT
        defw EXIT

;X .(		--	display comments
;   [CHAR] ) PARSE TYPE ; IMMEDIATE
    head DOTPAREN,2,".(",docolon,1,0
	defw LIT,$29,PARSE,TYPE,EXIT

; NUMERIC OUTPUT ================================
; Numeric conversion is done l.s.digit first, so
; the output buffer is built backwards in memory.

; Some double-precision arithmetic operators are
; needed to implement ANSI numeric conversion.

;Z UD/MOD   ud1 u2 -- u3 ud4   32/16->32 divide
;   >R 0 R@ UM/MOD  ROT ROT R> UM/MOD ROT ;
    head UDSLASHMOD,6,"UD/MOD",docolon,0,0
        defw TOR,ZERO,RFETCH,UMSLASHMOD,ROT,ROT
        defw RFROM,UMSLASHMOD,ROT,EXIT

;Z UD*      ud1 d2 -- ud3      32*16->32 multiply
;   DUP >R UM* DROP  SWAP R> UM* ROT + ;
    head UDSTAR,3,"UD*",docolon,0,0
        defw DUP,TOR,UMSTAR,DROP
        defw SWOP,RFROM,UMSTAR,ROT,PLUS,EXIT

;C HOLD  char --        add char to output string
;   -1 HP +!  HP @ C! ;
    head HOLD,4,"HOLD",docode,0,holdn-holdf
.holdf  ld hl,(usr_hp)
	dec hl
	ld (hl),c
	ld (usr_hp),hl
	pop bc
.holdn
	next

;C <#    --             begin numeric conversion
;   PAD HP ! ;          (initialize Hold Pointer)
    head LESSNUM,2,"<#",docode,0,lessnn-lessnf
.lessnf ld hl,usr_pad
	ld (usr_hp),hl
.lessnn
	next

;Z >DIGIT   n -- c      convert to 0..9A..Z
;   [ HEX ] DUP 9 > 7 AND + 30 + ;
     nohead TODIGIT,docode
	ld a,c
	cp 10
	jr c,todig1
	add a,7
.todig1 add a,$30
	ld c,a
	next

;C #     ud1 -- ud2     convert 1 digit of output
;   BASE @ UD/MOD ROT >DIGIT HOLD ;
    head NUM,1,"#",docolon,0,0
        defw BASE,FETCH,UDSLASHMOD,ROT,TODIGIT
        defw HOLD,EXIT

;C #S    ud1 -- ud2     convert remaining digits
;   BEGIN # 2DUP OR 0= UNTIL ;
    head NUMS,2,"#S",docolon,0,0
.NUMS1  defw NUM,TWODUP,OR,ZEROEQUAL,QBRANCH,NUMS1
        defw EXIT

;C #>    ud1 -- c-addr u    end conv., get string
;   2DROP HP @ PAD OVER - ;
    head NUMGREATER,2,"#>",docode,0,numgrn-numgrf
.numgrf pop bc
	ld bc,(usr_hp)
	push bc
	ld hl,usr_pad
	and a
	sbc hl,bc
	ld b,h
	ld c,l
.numgrn
	next

;C SIGN  n --           add minus sign if n<0
;   0< IF 2D HOLD THEN ;
    head SIGN,4,"SIGN",docolon,0,0
        defw ZEROLESS,QBRANCH,SIGN1,LIT,$2D,HOLD
.SIGN1  defw EXIT

;Z (U.)	u -- c-addr n
;   <# 0 #S #> ;
    nohead BRUDOT,docolon
        defw LESSNUM,ZERO,NUMS,NUMGREATER,EXIT

;C U.    u --           display u unsigned
;   (U.) TYPE SPACE ;
    head UDOT,2,"U.",docolon,0,0
        defw BRUDOT,TYPE,SPACE,EXIT

;Z (.)	n -- c-addr n
;   <# DUP ABS 0 #S ROT SIGN #> ;
    nohead BRDOT,docolon
        defw LESSNUM,DUP,ABS,ZERO,NUMS
        defw ROT,SIGN,NUMGREATER,EXIT

;C .     n --           display n signed
;   (.) TYPE SPACE ;
    head DOT,1,".",docolon,0,0
	defw BRDOT,TYPE,SPACE,EXIT

;Z (.R)	n1 c-addr n2 --
;   ROT OVER - 0 MAX SPACES TYPE ;
    nohead BRDOTR,docolon
	defw ROT,OVER,MINUS,ZERO,MAX,SPACES,TYPE,EXIT

;X U.R	u n --
;   SWAP (U.) (.R) ;
    head UDOTR,3,"U.R",docolon,0,0
	defw SWOP,BRUDOT,BRDOTR,EXIT

;X .R	n1 n2 --
;   SWAP (.) (.R) ;
    head DOTR,2,".R",docolon,0,0
	defw SWOP,BRDOT,BRDOTR,EXIT

;C DECIMAL  --      set number base to decimal
;   10 BASE ! ;
    head DECIMAL,7,"DECIMAL",docode,0,decimn-decimf
.decimf ld (iy+usr_base),10
	ld (iy+usr_base+1),0
.decimn
	next

;X HEX     --       set number base to hex
;   16 BASE ! ;
    head HEX,3,"HEX",docode,0,hexn-hexf
.hexf   ld (iy+usr_base),16
	ld (iy+usr_base+1),0
.hexn
	next

; DICTIONARY MANAGEMENT =========================

;C HERE    -- addr      returns dictionary ptr
;   DP @ ;
    head HERE,4,"HERE",docode,0,heren-heref
.heref  push bc
	ld hl,(usr_curds)
	ld c,(hl)
	inc hl
	ld b,(hl)
.heren
	next

;C ALLOT   n --         allocate n bytes in dict
;   DP +! ;
    head ALLOT,5,"ALLOT",docode,0,allotn-allotf
.allotf ld hl,(usr_curds)
	push de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl
	add hl,bc
	ex de,hl
	ld (hl),d
	dec hl
	ld (hl),e
	pop de
	pop bc
.allotn
	next

; Note: , and C, are only valid for combined
; Code and Data spaces.

;C ,    x --           append cell to dict
;   HERE ! 1 CELLS ALLOT ;
    head COMMA,1,$2c,docode,0,comman-commaf
.COMMADEST
.commaf ld hl,(usr_curds)
	push de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl
	ld (hl),c
	inc hl
	ld (hl),b
	inc hl
	ex de,hl
	ld (hl),d
	dec hl
	ld (hl),e
	pop de
	pop bc
.comman
	next

;C C,   char --        append char to dict
;   HERE C! 1 CHARS ALLOT ;
    head CCOMMA,2,"C"&$2c,docode,0,ccommn-ccommf
.ccommf ld hl,(usr_curds)
	push de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl
	ld (hl),c
	inc hl
	ex de,hl
	ld (hl),d
	dec hl
	ld (hl),e
	pop de
	pop bc
.ccommn
	next


; WORDLISTS =====================================

;Z NEXTLIST	-- a-addr	next available wid
    nohead NEXTLIST,docon
	defw usr_nextlist

;Z SORDER	-- a-addr	search-order specification
    nohead SORDER,docon
	defw usr_sorder

;X FORTH-WORDLIST	-- wid	wordlist ID for all FORTH words
    head FORTHWORDLIST,14,"FORTH-WORDLIST",docon,0,0
	defw usr_forth

;X GET-CURRENT		-- wid	get compilation wid
;   CURRENT @ ;
    head GETCURRENT,11,"GET-CURRENT",docode,0,getcun-getcuf
.getcuf push bc
	ld bc,(usr_current)
.getcun
	next

;X SET-CURRENT		wid --	set compilation wid
;   CURRENT ! ;
    head SETCURRENT,11,"SET-CURRENT",docode,0,setcun-setcuf
.setcuf ld (usr_current),bc
	pop bc
.setcun
	next

;X WORDLIST		-- wid	get new wid
;   NEXTLIST @
;   1 CELLS NEXTLIST +!
;   0 OVER ! ;
    head WORDLIST,8,"WORDLIST",docolon,0,0
	defw NEXTLIST,FETCH
	defw ONE,CELLS,NEXTLIST,PLUSSTORE
	defw ZERO,OVER,STORE,EXIT

;X DEFINITIONS		--	set current to first in search order
;   SORDER @ IF  SORDER CELL+ @ SET-CURRENT  THEN ;
    head DEFINITIONS,11,"DEFINITIONS",docolon,0,0
	defw SORDER,FETCH,QBRANCH,defs1
	defw SORDER,CELLPLUS,FETCH,SETCURRENT
.defs1  defw EXIT

;X GET-ORDER		-- widn...wid1 n	get search order
;   SORDER @ DUP IF
;     1 SWAP DO  SORDER I CELLS + @  -1 +LOOP
;     SORDER @
;   THEN ;
    head GETORDER,9,"GET-ORDER",docolon,0,0
	defw SORDER,FETCH,DUP,QBRANCH,getor2
	defw ONE,SWOP,XDO
.getor1 defw SORDER,II,CELLS,PLUS,FETCH
	defw MINUSONE,XPLUSLOOP,getor1
	defw SORDER,FETCH
.getor2 defw EXIT

;X SET-ORDER		widn...wid1 n --	set search order
;   CASE  -1 OF  FORTH-WORDLIST 1 RECURSE  ENDOF
;          0 OF  0 SORDER !  ENDOF
;   DUP >R
;   8 U> IF  -49 THROW  THEN
;   R@ SORDER ! SORDER
;   R@ FOR  CELL+ SWAP OVER !  STEP  DROP
;   R>
;   ENDCASE ;
    head SETORDER,9,"SET-ORDER",docolon,0,0
	defw TRUE,QBRANCH,setor1
	defw MINUSONE,OVER,EQUAL,QBRANCH,setor2,DROP
	defw FORTHWORDLIST,ONE,SETORDER
.setor1 defw EXIT
.setor2 defw ZERO,OVER,EQUAL,QBRANCH,setor4,DROP
	defw ZERO,SORDER,STORE
.setor3 defw EXIT
.setor4 defw DUP,TOR
	defw LIT,8,UGREATER,QBRANCH,setor5
	defw LIT,-49,THROW
.setor5 defw RFETCH,SORDER,STORE,SORDER
	defw RFETCH,TOR
.setor6 defw CELLPLUS,SWOP,OVER,STORE,BRSTEP,setor6
	defw DROP,RFROM
	defw DROP
.setor7 defw EXIT

;X ONLY		--	set minimum search order
;   -1 SET-ORDER ;
    head ONLY,4,"ONLY",docolon,0,0
	defw MINUSONE,SETORDER,EXIT

;X ALSO		--	duplicate first list in search order
;   GET-ORDER OVER SWAP 1+ SET-ORDER ;
    head ALSO,4,"ALSO",docolon,0,0
	defw GETORDER,OVER,SWOP,ONEPLUS,SETORDER,EXIT

;X PREVIOUS	--	remove first list in search order
;   GET-ORDER ?DUP IF  NIP 1- SET-ORDER  THEN ;
    head PREVIOUS,8,"PREVIOUS",docolon,0,0
	defw GETORDER,QDUP,QBRANCH,prev1
	defw NIP,ONEMINUS,SETORDER
.prev1  defw EXIT

;X ORDER	--	display search order and compilation wid
;   GET-ORDER ?DUP
;   IF  FOR  CR U.  STEP  THEN
;   GET-CURRENT U. ;
    head ORDER,5,"ORDER",docolon,0,0
	defw GETORDER,QDUP,QBRANCH,order2
	defw TOR
.order1 defw CRR,UDOT,BRSTEP,order1
.order2	defw GETCURRENT,UDOT,EXIT

;X SEARCH-WORDLIST	caddr u wid -- 0 | xt 1 | xt -1
;   BEGIN  @ ?DUP
;   WHILE  2DUP C@ =			\ -- caddr u nfa flag
;          IF  DUP 1+ 2OVER S= 0=	\ -- caddr u nfa flag
;              IF  NIP NIP DUP NFA>CFA	\ -- nfa xt
;                  SWAP IMMED? 0= 1 OR  \ -- xt 1 | xt -1
;                  EXIT
;              THEN
;          THEN
;          NFA>LFA			\ -- caddr u lfa
;   REPEAT  2DROP FALSE ;
    head SEARCHWORDLIST,15,"SEARCH-WORDLIST",docode,0,0
	push bc
	exx
	pop hl				; HL=wid
	pop bc				; C=u
.swidlp	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl			; HL=nfa
	ld a,h
	or l
	jr z,swid1			; exit if null
	ld a,(hl)
	cp c
	jr nz,swid2			; skip if counts not equal
	pop de
	push de				; de=caddr
	push hl				; save nfa
	ld b,c				; b=count
	inc hl
.swideq	ld a,(de)
	cp (hl)				; check for equal strings
	jr nz,swid3
	inc de
	inc hl
	djnz swideq
	pop de				; discard nfa
	pop de				; discard c-addr
	ld a,(hl)			; A=immed/fast byte
	inc hl
	push hl				; stack xt
	exx
	ld bc,-1			; most words are normal
	rla
	jp nc,next_code
	inc bc
	inc bc				; TOS=1 if immediate
	jp next_code
.swid3	pop hl				; restore nfa
.swid2	dec hl				; nfa>lfa
	dec hl
	jp swidlp
.swid1	exx
	pop bc				; discard caddr
	ld bc,0				; false flag
	next

;Z VOCABULARY	"<spaces>name" --	create new vocabulary
;   NSDS WORDLIST CREATE ,
;   DOES> @ >R
;         GET-ORDER NIP R> SWAP SET-ORDER ;
    head VOCABULARY,10,"VOCABULARY",docolon,0,0
	defw NSDS,WORDLIST,CREATE,COMMA
        defw XDOES
.dovoc  call dodoes
	defw FETCH,TOR
	defw GETORDER,NIP,RFROM,SWOP,SETORDER,EXIT

;Z FORTH	--	set first list in search order to FORTH-WORDLIST
;   VOCABULARY FORTH  FORTH-WORDLIST ' FORTH >BODY !
    head FORTH,5,"FORTH",dovoc,0,0
	defw usr_forth

; INTERPRETER ===================================
; Note that NFA>LFA, NFA>CFA, IMMED?, and FIND
; are dependent on the structure of the Forth
; header.  This may be common across many CPUs,
; or it may be different.

;C SOURCE   -- adr n    current input buffer
;   'SOURCE 2@ ;        length is at lower adrs
    head SOURCE,6,"SOURCE",docolon,0,0
        defw TICKSOURCE,TWOFETCH,EXIT

;X /STRING  a u n -- a+n u-n   trim string
;   ROT OVER + ROT ROT - ;
    head SLASHSTRING,7,"/STRING",docode,0,slstrn-slstrf
.slstrf pop hl
	and a
	sbc hl,bc	; u-n
	ex (sp),hl
	add hl,bc	; a+n
	ex (sp),hl
	ld b,h
	ld c,l
.slstrn
	next

;X PARSE	char -- c-addr n	parse delimited by char
;   SOURCE >IN @ /STRING
;   2DUP 2>R ROT SCAN
;   DUP IF  CHAR-  THEN
;   2R> ROT - >IN +!
;   TUCK - ;
    head PARSE,5,"PARSE",docode,0,0
	ld a,c			; A=char
	exx
	ld de,(usr_in)		; DE=>IN @
	ld hl,(usr_source)
	and a
	sbc hl,de
	ld b,h
	ld c,l			; BC=len (after /STRING)
	ld hl,(usr_source+2)
	add hl,de		; HL=c-addr (after /STRING)
	ld e,a			; E=char
	push hl
	push bc
	cp $20
	jr nz,parse5		; special case for BLs+control chars follows
.parse4	ld a,b
	or c
	jr z,parse7
	ld a,(hl)
	cp $21
	dec bc
	jr c,parse7		; done scan BLs+control chars
	inc hl
	jp parse4
.parse5	ld a,b
	or c
	jr z,parse7
	ld a,(hl)
	cp e
	dec bc
	jr z,parse7		; done scan matching chars
	inc hl
	jp parse5
.parse7	ex (sp),hl		; TOS=c-addr', HL=n, BC=n'
	and a
	sbc hl,bc
	ld bc,(usr_in)
	add hl,bc
	ld (usr_in),hl
	exx
	pop hl
	pop bc
	push bc			; HL=addr', 2OS=BC=addr
	and a
	sbc hl,bc
	ld b,h
	ld c,l			; TOS=N
	next


;C WORD   char -- c-addr   word delim'd by char, skip leading delimiters
;   DUP  SOURCE >IN @ /STRING   -- c c adr n
;   DUP >R   ROT SKIP           -- c adr' n'
;   OVER >R  ROT SCAN           -- adr" n"
;   DUP IF CHAR- THEN        skip trailing delim.
;   R> R> ROT -   >IN +!        update >IN offset
;   TUCK -                      -- adr' N
;   HERE >counted               --
;   HERE                        -- a
;   BL OVER COUNT + C! ;    append trailing blank
    head WORD,4,"WORD",docode,0,0
    	ld a,c			; A=char
	exx
	ld de,(usr_in)		; DE=>IN @
	ld hl,(usr_source)
	and a
	sbc hl,de
	ld b,h
	ld c,l			; BC=len (after /STRING)
	ld hl,(usr_source+2)
	add hl,de		; HL=c-addr (after /STRING)
	ld e,a			' E=char
	push bc			; save initial length
	cp $20
	jr nz,word3		; special case for BLs+control chars follows
.word2	ld a,b
	or c
	jr z,word1
	ld a,(hl)
	cp $21
	jr nc,word1		; done skip BLs+control chars
	inc hl
	dec bc
	jp word2
.word3	ld a,b
	or c
	jr z,word1
	ld a,(hl)
	cp e
	jr nz,word1		; done skip matching chars
	inc hl
	dec bc
	jp word3
.word1	push hl			; save address of word start
	ld a,e
	cp $20
	jr nz,word5		; special case for BLs+control chars follows
.word4	ld a,b
	or c
	jr z,word6
	ld a,(hl)
	cp $21
	dec bc
	jr c,word6		; done scan BLs+control chars
	inc hl
	jp word4
.word5	ld a,b
	or c
	jr z,word6
	ld a,(hl)
	cp e
	dec bc
	jr z,word6		; done scan matching chars
	inc hl
	jp word5
.word6	pop de			; HL=addr", BC=n", DE=addr'
	ex (sp),hl		; HL=n
	and a
	sbc hl,bc		; HL=n-n"
	ld bc,(usr_in)
	add hl,bc
	ld (usr_in),hl		; update >IN
	pop hl
	and a
	sbc hl,de		; HL=addr"-addr'=N; DE=addr'
	ld a,l			; A=N
	ld hl,(usr_curds)
	push de
	ld e,(hl)
	inc hl
	ld d,(hl)		; DE=HERE
	pop hl			; HL=addr'
	ld (de),a		; store count
	push de			; save HERE
	inc de
	and a
	jr z,word7		; skip CMOVE if no count
	ld c,a
	ld b,0
	ldir			; copy string to HERE
.word7	ld a,$20
	ld (de),a		; append a BL
	exx
	pop bc			; TOS=address of string
	next

;Z NFA>CFA   nfa -- cfa    name adr -> code field
;   COUNT 7F AND + 1+ ;       mask off 'smudge' bit
    nohead NFATOCFA,docode
	ld a,(bc)
	inc bc
	and $7F
	add a,c
	ld c,a
	jr nc,nfacf1
	inc b
.nfacf1 inc bc
	next

;C FIND   c-addr -- c-addr 0   if not found
;C                  xt  1      if immediate
;C                  xt -1      if "normal"
;   FALSE SORDER @ ?DUP		\ -- caddr 0 flag
;   IF  0 DO  OVER COUNT	\ -- caddr 0 caddr+1 u
;	      SORDER I 1+ CELLS + @	\ -- caddr 0 caddr+1 u wid
;	      SEARCH-WORDLIST	\ -- caddr 0; 0 | xt 1 | xt -1
;	      ?DUP IF  2SWAP 2DROP UNLOOP EXIT  THEN  \ xt 1 | xt -1
;	  LOOP
;   THEN ;			\ -- caddr 0
    head FIND,4,"FIND",docolon,0,0
	defw FIND1
.find1	ld hl,usr_sorder	; start of ORDER
	ld a,(hl)		; max 8 wordlists in order, so only need 1 byte!
.findlp	and a
	jr z,find2		; end if no more wordlists
	dec a
	push af			; save wordlist details
	inc hl
	inc hl
	push hl
	push bc			; save c-addr
	ld a,(bc)
	inc bc
	push bc			; stack c-addr'
	ld c,a
	ld b,0
	push bc			; stack u
	ld c,(hl)
	inc hl
	ld b,(hl)		; TOS=wid
	ld de,find3
	jp SEARCHWORDLIST
.find3	defw find3+2
	ld a,c
	and a
	jr nz,find4		; move on if found
	pop bc			; restore c-addr & wordlist details
	pop hl
	pop af
	jr findlp
.find2	push bc			; not found, so leave c-addr 0
	ld bc,0
	jp EXIT
.find4	pop hl			; hold xt
	pop af			; discard c-addr & wordlist details
	pop af
	pop af
	push hl			; restack xt
	jp EXIT			; leave xt 1 | xt -1 from SEARCH-WORDLIST

;C LITERAL  x --        append numeric literal
;   STATE C@ IF ,LIT THEN ; IMMEDIATE
; This tests STATE so that it can also be used
; interpretively.  (ANSI doesn't require this.)
    head LITERAL,7,"LITERAL",docode,1,0
	ld a,(usr_state)
	and a
	jp z,next_code			; do nothing if interpreting
	; continue straight into ,LIT

; NOTE: ,LIT must be headerless code and follow on directly from LITERAL!!
;Z ,LIT		x --	compile numeric literal into colon definition
;   FAST?
;   IF  $01C5		; push bc / ld bc,
;   ELSE  ['] LIT
;   THEN  , , ;
    nohead COMMALIT,docode
	ld a,(usr_fastcomp)
	push bc
	exx
	ld bc,LIT
	and a
	jr z,comlt1
	ld bc,$01c5
.comlt1	ld hl,(usr_curds)		; HL=address of DP
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl			; DE=address of DP+1, HL=DP
	ld (hl),c
	inc hl
	ld (hl),b			; first comma
	inc hl
	pop bc
	ld (hl),c
	inc hl
	ld (hl),b			; second comma
	inc hl
	ex de,hl
	ld (hl),d			; store the new DP
	dec hl
	ld (hl),e
	exx
	pop bc				; get new TOS
	next

;C >NUMBER  ud adr u -- ud' adr' u'
;C                      convert string to number
;   BEGIN
;   DUP WHILE
;       OVER C@ DIGIT?
;       0= IF EXIT THEN
;       >R 2SWAP BASE @ UD*
;       R> M+ 2SWAP
;       1 /STRING
;   REPEAT ;
    head TONUMBER,7,">NUMBER",docode,0,0
	pop hl			; HL=addr, BC=u
	exx
	pop de
	pop hl			; D'E'H'L'=ud
	exx
.tonum2	ld a,b
	or c
	jr z,tonum1
	ld a,(hl)
	sub '0'
	jr c,tonum1
	cp 10
	jr c,okdig
	sub 7
	jr c,tonum1
.okdig	exx
	ld c,a
	ld a,(iy+usr_base)	; assumes base<256 (acceptable)
	dec a
	cp c
	exx
	jr c,tonum1
	exx
	push bc			; save C=digit to add in
	cp 9
	jr z,decnum		; go to quick decimal *+
	cp 15
	jr z,hexnum		; go to quick hex *+
	ex de,hl
	push de
	ex (sp),iy
	ld b,h
	ld c,l			; HLIY=BCDE=ud, A=base-1
.tonum3	add iy,de
	adc hl,bc
	dec a
	jr nz,tonum3		; HLIY=ud*base
	ex de,hl
	ex (sp),iy
	pop hl			; DEHL=ud*base
	jr addin
.hexnum add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*2
	add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*4
	add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*8
	add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*16
	jr addin
.decnum	add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*2
	push hl
	ex (sp),iy
	ld b,d
	ld c,e			; BCIY=DEHL=ud*2
	add hl,hl
	ex de,hl
	adc hl,hl
	ex de,hl		; DEHL=ud*4
	add hl,hl
	ex de,hl
	adc hl,hl		; HLDE=ud*8
	add iy,de
	adc hl,bc		; HLIY=ud*10
	ex de,hl
	ex (sp),iy
	pop hl			; DEHL=ud*10
.addin	pop bc
	ld b,0
	add hl,bc
	ex de,hl
	ld c,0
	adc hl,bc
	ex de,hl		; DEHL=ud*BASE+digit
	exx
	inc hl			; 1 /STRING
	dec bc
	jr tonum2
.tonum1	exx
	push hl
	push de			; restack ud
	exx
	push hl			; restack address
	next

;Z INTERPRET    i*x -- j*x
;Z                      interpret given buffer
; This is a common factor of EVALUATE and QUIT.
; ref. dpANS-6, 3.4 The Forth Text Interpreter
;   0 >IN !
;   BEGIN
;   BL WORD DUP C@ WHILE        -- textadr
;       FIND                    -- a 0/1/-1
;       ?DUP IF                 -- xt 1/-1
;           1+ STATE C@ 0= OR    immed or interp?
;           IF EXECUTE ELSE ,XT THEN
;       ELSE                    -- textadr
;           ?NUMBER
;           IF POSTPONE LITERAL     converted ok
;           ELSE -13 THROW	undefined word exception
;           THEN
;       THEN
;   REPEAT DROP ;
    nohead INTERPRET,docolon
        defw INTER1
.INTER1	ld hl,0
	ld (usr_in),hl
.INTLOP	push bc
	ld c,$20
	ld de,INTER2
	jp WORD
.INTER2 defw INTER2+2
	ld a,(bc)
	and a
	jr z,INTER3		; go to exit if null string
	ld de,INTER4
	jp FIND
.INTER4	defw INTER4+2
	ld a,c
	pop bc			; BC=xt or c-addr
	and a
	jr z,INTER5		; not found, so try to find number
	ld de,INTER8		; set IP to loop address
	jp p,EXECUTE		; immediate, so execute and loop back
	ld a,(usr_state)
	and a
	jp nz,COMMAXT		; compile and loop back if compiling
	jp EXECUTE		; execute and loop back if interpreting

; This section used to be a separate word called ?NUMBER
.INTER5
	ld hl,0
	push hl
	push hl			; stack D0
	ld h,b
	ld l,c
	ld c,(hl)
	ld b,0
	inc hl			; HL=c-addr, BC=count
	ld a,c
	and a
	scf
	jr z,qnum2		; skip this bit if no length
	ld a,(hl)
	cp '+'
	scf
	jr z,qnplus
	cp '-'
	scf
	jr nz,qnum2
	and a			; Fc=0 for negative
.qnplus	inc hl			; skip +/- sign
	dec bc
.qnum2	ex af,af'		; Fc'=sign; won't be corrupted by NEXT/>NUMBER
	push hl			; stack c-addr
	ld de,QNUM3
	jp TONUMBER
.qnum3	defw QNUM3+2
	ld a,b
	or c
	pop bc
	pop bc
	pop bc			; TOS=number
	jr nz,INTER7		; error if string not exhausted
	ex af,af'
	jr c,qnum4
	ld a,b
	cpl			; negate
	ld b,a
	ld a,c
	cpl
	ld c,a
	inc bc
.qnum4	ld de,INTER8
	jp LITERAL		; leave on stack/compile and loop back
.INTER7	ld bc,-13
	jp THROW
.INTER3	pop bc
	jp EXIT
.INTER8	defw INTLOP


;Z NEST-INPUT	xt caddr u id blk --
;   SAVE-INPUT DUP
;   BEGIN  ROT >R 1- ?DUP 0= UNTIL >R	; save current input specification
;   BLK !  'SOURCE-ID !  'SOURCE 2!	; set BLK, SOURCE-ID and SOURCE
;   CATCH				; interpret single/multiple
;   R> DUP
;   BEGIN  R> ROT ROT 1- ?DUP 0=  UNTIL
;   RESTORE-INPUT			; restore old input specification
;   IF  -4094 THROW  THEN		; exception -4094 if unable to
;   THROW ;				; reTHROW any interpretation exception
    nohead NESTINPUT,docolon
	defw SAVEINPUT,DUP
.ninpt2	defw ROT,TOR,ONEMINUS,QDUP,ZEROEQUAL,QBRANCH,ninpt2,TOR
	defw BLK,STORE,TICKSOURCEID,STORE,TICKSOURCE,TWOSTORE
	defw CATCH
	defw RFROM,DUP
.ninpt3	defw RFROM,ROT,ROT,ONEMINUS,QDUP,ZEROEQUAL,QBRANCH,ninpt3
	defw RESTOREINPUT
	defw QBRANCH,ninpt1
	defw LIT,-4094,THROW
.ninpt1 defw THROW,EXIT


;C EVALUATE  i*x c-addr u -- j*x  interprt string
;   ['] INTERPRET  ROT ROT -1 0  NEST-INPUT ;
    head EVALUATE,8,"EVALUATE",docolon,0,0
        defw LIT,INTERPRET,ROT,ROT,MINUSONE,ZERO
	defw NESTINPUT,EXIT

;Z INTERPRETK		--	interpret from keyboard
;   CR 0 BLK !  0 'SOURCE-ID !
;   REFILL DROP			\ always TRUE with keyboard source
;   INTERPRET ;
    nohead INTERPRETK,docolon
	defw CRR,ZERO,BLK,STORE
	defw ZERO,TICKSOURCEID,STORE
	defw REFILL,DROP,INTERPRET,EXIT


;C QUIT     --    R: i*x --    outer interpreter
;   R0 RP!  0 STATE ! 0 usr_fastcomp C!
;   BEGIN  ['] INTERPRETK CATCH ?DUP
;          IF  CASE  -1 OF  ENDOF
;                    -2 OF  ABORT"S @ ABORT"S 1+ 1+ @ TYPE  ENDOF
;                    DUP ." Exception # " .
;	       ENDCASE
;              REVEAL 0 STATE ! 0 usr_fastcomp C! FORTH
;          ELSE  STATE C@ 0= IF  ."  ok"  THEN
;          THEN
;   AGAIN ;
    head QUIT,4,"QUIT",docolon,0,0
        defw R0,RPSTORE
.QUIT2	defw ZERO,STATE,STORE
	defw ZERO,LIT,usr_fastcomp,CSTORE
.QUIT1  defw LIT,INTERPRETK,CATCH,QDUP
        defw QBRANCH,QUIT11
	defw TRUE,QBRANCH,QUIT5
	defw MINUSONE,OVER,EQUAL,QBRANCH,QUIT6,DROP
.QUIT5  defw BRANCH,QUIT8
.QUIT6  defw LIT,-2,OVER,EQUAL,QBRANCH,QUIT10,DROP
        defw ABORTQS,FETCH,ABORTQS,ONEPLUS,ONEPLUS,FETCH,TYPE
.QUIT9  defw BRANCH,QUIT8
.QUIT10 defw DUP,XSQUOTE
	defm 12&"Exception # "
        defw TYPE,DOT
	defw DROP
.QUIT8  defw REVEAL,FORTH,BRANCH,QUIT2
.QUIT11 defw STATE,CFETCH,ZEROEQUAL,QBRANCH,QUIT1
	defw XSQUOTE
        defm 3&" ok"
        defw TYPE
.QUIT7  defw BRANCH,QUIT1         ; eternal loop

;Z HANDLER    -- addr
    nohead HANDLER,docode
    	push bc
	push iy
	pop hl
	ld bc,usr_handler
	add hl,bc
	ld b,h
	ld c,l
	next

;X CATCH     i*x xt -- j*x 0 | i*x n
;   SP@ >R  LP @ >R  HANDLER @ >R
;   RP@ HANDLER !
;   EXECUTE
;   R> HANDLER !  2R> 2DROP
;   0 ;   
    head CATCH,5,"CATCH",docolon,0,0
	defw SPFETCH,TOR,LP,FETCH,TOR,HANDLER,FETCH,TOR
	defw RPFETCH,HANDLER,STORE
	defw EXECUTE
	defw RFROM,HANDLER,STORE,TWORFROM,TWODROP
	defw ZERO,EXIT

;X THROW     k*x n -- k*x | i*x n
;   ?DUP IF  HANDLER @ RP!  R> HANDLER !
;	     R> LP !
;	     R> SWAP >R
;	     SP! DROP R>
;        THEN ;
    head THROW,5,"THROW",docolon,0,0
	defw QDUP,QBRANCH,THROW2
	defw HANDLER,FETCH,RPSTORE,RFROM,HANDLER,STORE
	defw RFROM,LP,STORE
	defw RFROM,SWOP,TOR
	defw SPSTORE,DROP,RFROM
.THROW2	defw EXIT


;C ABORT    i*x --   R: j*x --   As per EXCEPTION EXT
;   -1 THROW ;
    head ABORT,5,"ABORT",docolon,0,0
        defw MINUSONE,THROW	; THROW never returns

;Z ABORT"S    -- addr
    head ABORTQS,7,"ABORT"&$22&"S",docode,0,abrtqn-abrtqf
.abrtqf	push bc
	push iy
	pop hl
	ld bc,usr_abortq
	add hl,bc
	ld b,h
	ld c,l
.abrtqn
	next

;Z ?ABORT   f c-addr u --      abort & print msg
;   ABORT"S 1+ 1+ ! ABORT"S ! IF  -2 THROW  THEN ;
    nohead QABORT,docolon
	defw ABORTQS,ONEPLUS,ONEPLUS,STORE,ABORTQS,STORE
	defw QBRANCH,QABO2
	defw LIT,-2,THROW
.QABO2  defw EXIT

;C ABORT"  i*x 0  -- i*x   R: j*x -- j*x  x1=0
;C         i*x x1 --       R: j*x --      x1<>0
;   POSTPONE S" POSTPONE ?ABORT ; IMMEDIATE
    head ABORTQUOTE,6,"ABORT"&$22,docolon,1,0
        defw SQUOTE
        defw LIT,QABORT,COMMAXT
        defw EXIT

;C '    -- xt           find word in dictionary
;   BL WORD FIND
;   0= IF  -13 THROW  THEN ;
    head TICK,1,"'",docolon,0,0
	defw BL,WORD,FIND,ZEROEQUAL
	defw QBRANCH,tick2,LIT,-13,THROW
.tick2	defw EXIT

;C CHAR   -- char           parse ASCII character
;   BL WORD 1+ C@ ;
    head CHAR,4,"CHAR",docolon,0,0
        defw BL,WORD,ONEPLUS,CFETCH,EXIT

;C [CHAR]   --          compile character literal
;   CHAR  ,LIT ; IMMEDIATE
    head BRACCHAR,6,"[CHAR]",docolon,1,0
        defw CHAR,COMMALIT,EXIT

; COMPILER ======================================

;Z NS	--		set code/namespace to current dataspace
;   CURDS @ CURNS ! ;	
    head NS,2,"NS",docode,0,nspacn-nspacf
.nspacf ld hl,(usr_curds)
	ld (usr_curns),hl
.nspacn
	next

;Z NSDS	--		ensure namespace=dataspace
;			exception -4093 thrown if not
    nohead NSDS,docode
	exx
	ld hl,(usr_curds)
	ld de,(usr_curns)
	and a
	sbc hl,de
	exx
	jp z,next_code		; okay if namespace=dataspace
	ld bc,-4093
	jp THROW		; exception -4093 if not the same

;X :NONAME	-- xt		create unnamed colon definition
;   NSDS
;   HERE docolon ,CF		get execution token, setup code field
;   1 STATE ! ;
    head NONAME,7,":NONAME",docolon,0,0
	defw NSDS,HERE,LIT,docolon,COMMACF
	defw ONE,STATE,STORE,EXIT	

;C CREATE   --      create an empty definition
;   CURDS @ CURNS @ CURDS !    save dataspace, set to namespace
;   GET-CURRENT @ ,		link field
;   HERE GET-CURRENT !		update compilation wordlist link
;   HERE LATEST !               new "latest" link
;   BL WORD C@ 1+ ALLOT         name field
;   0 C,			immediate/fast field
;   docreate ,CF                code field
;   HERE 1 CELLS ALLOT		reserve space for pointer to data field
;   SWAP CURDS !		restore dataspace
;   HERE SWAP !	;		store data field address
    head CREATE,6,"CREATE",docolon,0,0
	defw CURDS,FETCH,CURNS,FETCH,CURDS,STORE
        defw GETCURRENT,FETCH,COMMA
	defw HERE,GETCURRENT,STORE
        defw HERE,LATEST,STORE
        defw BL,WORD,CFETCH,ONEPLUS,ALLOT
	defw ZERO,CCOMMA
        defw LIT,docreate,COMMACF
	defw HERE,ONE,CELLS,ALLOT
	defw SWOP,CURDS,STORE
	defw HERE,SWOP,STORE,EXIT

;C VARIABLE	--	create a variable
;   CURDS @ RAM			; save current dataspace, change to RAM
;   CREATE 1 CELLS ALLOT
;   CURDS ! ;			; restore dataspace
    head VARIABLE,8,"VARIABLE",docolon,0,0
	defw CURDS,FETCH,RAM
	defw CREATE,ONE,CELLS,ALLOT
	defw CURDS,STORE,EXIT

;Z (DOES>)  --      run-time action of DOES>
;   R>              adrs of headless DOES> def'n
;   LATEST @ NFA>CFA    code field to fix up
;   !CF ;
    nohead XDOES,docolon
        defw RFROM,LATEST,FETCH,NFATOCFA,STORECF
        defw EXIT

;C DOES>    --      change action of latest def'n
;   COMPILE (DOES>)
;   dodoes ,CF ; IMMEDIATE
    head DOES,5,"DOES>",docolon,1,0
        defw LIT,XDOES,COMMAXT
        defw LIT,dodoes,COMMACF,EXIT

;C RECURSE  --      recurse current definition
;   LATEST @ NFA>CFA ,XT ; IMMEDIATE
    head RECURSE,7,"RECURSE",docolon,1,0
        defw LATEST,FETCH,NFATOCFA,COMMAXT,EXIT

;C [        --      enter interpretive state
;   0 STATE C! ; IMMEDIATE
    head LEFTBRACKET,1,"[",docode,1,leftbn-leftbf
.leftbf xor a
	ld (usr_state),a
.leftbn
	next

;C ]        --      enter compiling state
;   1 STATE C! ;
    head RIGHTBRACKET,1,"]",docode,0,rigtbn-rigtbf
.rigtbf ld a,1
	ld (usr_state),a
.rigtbn
	next

;Z HIDE     --      "hide" latest definition
;   LATEST @ DUP C@ 80 OR SWAP C! ;
    nohead HIDE,docode
      ld hl,(usr_latest)
      set 7,(hl)
      next

;Z REVEAL   --      "reveal" latest definition
;   LATEST @ DUP C@ 7F AND SWAP C! ;
    nohead REVEAL,docode
      ld hl,(usr_latest)
      res 7,(hl)
      next

;C :        --      begin a colon definition
;   NSDS CREATE HIDE 1 STATE ! !COLON ;
    head COLON,1,":",docolon,0,0
        defw NSDS,CREATE,HIDE,ONE,STATE,STORE,STORCOLON
        defw EXIT

;C ;
;  REVEAL
;  FAST? IF  0C3 C,  THEN
;  ['] EXIT ,  0 STATE ! 0 usr_fastcomp C! ;  IMMEDIATE
    head SEMICOLON,1,";",docolon,1,0
        defw REVEAL,FASTQ
	defw QBRANCH,SEMI2
	defw LIT,$c3,CCOMMA
.SEMI2	defw LIT,EXIT,COMMA
        defw ZERO,STATE,STORE,ZERO,LIT,usr_fastcomp,CSTORE,EXIT

;C [']  --         find word & compile as literal
;   '  ,LIT ; IMMEDIATE
; When encountered in a colon definition, the
; phrase  ['] xxx  will cause   LIT,xxt  to be
; compiled into the colon definition (where
; (where xxt is the execution token of word xxx).
; When the colon definition executes, xxt will
; be put on the stack.  (All xt's are one cell.)
    head BRACTICK,3,"[']",docolon,1,0
	defw TICK,COMMALIT,EXIT

;C POSTPONE  --   postpone compile action of word
;   BL WORD FIND DUP 0= IF  -13 THROW  THEN
;   0< IF   -- xt  non immed: add code to current
;                  def'n to compile xt later.
;       ['] LIT ,XT  ,      add "LIT,xt,COMMAXT"
;       ['] ,XT ,XT         to current definition
;   ELSE  ,XT      immed: compile into cur. def'n
;   THEN ; IMMEDIATE
    head POSTPONE,8,"POSTPONE",docolon,1,0
	defw BL,WORD,FIND,DUP,ZEROEQUAL,QBRANCH,POST3
	defw LIT,-13,THROW
.POST3  defw ZEROLESS,QBRANCH,POST1
        defw LIT,LIT,COMMAXT,COMMA
        defw LIT,COMMAXT
.POST1  defw COMMAXT
.POST2  defw EXIT


; CONTROL STRUCTURES ============================

;C IF       -- adrs    conditional forward branch
;   FAST?
;   IF  compile code: ld a,b; or c; pop bc; jp z,
;   ELSE  ['] qbranch ,BRANCH
;   THEN
;   HERE DUP ,DEST ;
;   IMMEDIATE
    head IF,2,"IF",docolon,1,0
	defw FASTQ
	defw QBRANCH,IF1
	defw LIT,$b178,COMMA,LIT,$cac1,COMMA
	defw BRANCH,IF2
.IF1	defw LIT,QBRANCH,COMMABRANCH
.IF2	defw HERE,DUP,COMMADEST,EXIT

;Z ?IF      -- adrs   non-destructive conditional forward branch
;   FAST?
;   IF  compile code: ld a,b;or c;jp z,
;   ELSE  ['] DUP ,XT  ['] qbranch ,BRANCH
;   THEN
;   HERE DUP ,DEST ; IMMEDIATE
    head QUERYIF,3,"?IF",docolon,1,0
	defw FASTQ
	defw QBRANCH,QIF1
	defw LIT,$b178,COMMA,LIT,$ca,CCOMMA
	defw BRANCH,QIF2
.QIF1	defw LIT,DUP,COMMAXT,LIT,QBRANCH,COMMABRANCH
.QIF2	defw HERE,DUP,COMMADEST,EXIT

;C THEN     adrs --        resolve forward branch
;   HERE SWAP !DEST ; IMMEDIATE
    head THEN,4,"THEN",docolon,1,0
        defw HERE,SWOP,STOREDEST,EXIT

;C ELSE     adrs1 -- adrs2    branch for IF..ELSE
;   FAST?
;   IF  compile code: jp
;   ELSE  ['] branch ,BRANCH
;   THEN
;   HERE DUP ,DEST
;   SWAP  POSTPONE THEN ; IMMEDIATE
    head ELSE,4,"ELSE",docolon,1,0
	defw FASTQ
	defw QBRANCH,ELSE1
	defw LIT,$c3,CCOMMA
	defw BRANCH,ELSE2
.ELSE1	defw LIT,BRANCH,COMMABRANCH
.ELSE2	defw HERE,DUP,COMMADEST
        defw SWOP,THEN,EXIT

;C BEGIN    -- adrs        target for bwd. branch
;   HERE ; IMMEDIATE
    head BEGIN,5,"BEGIN",docode,1,0
        jp HERE

;C UNTIL    adrs --   conditional backward branch
;   FAST?
;   IF  compile code: ld a,b; or c; pop bc; jp z,
;   ELSE  ['] qbranch ,BRANCH
;   THEN  ,DEST ; IMMEDIATE
;   conditional backward branch
    head UNTIL,5,"UNTIL",docolon,1,0
	defw FASTQ
	defw QBRANCH,UNTIL1
	defw LIT,$b178,COMMA,LIT,$cac1,COMMA
	defw BRANCH,UNTIL2
.UNTIL1	defw LIT,QBRANCH,COMMABRANCH
.UNTIL2	defw COMMADEST,EXIT

;X AGAIN    adrs --      uncond'l backward branch
;   FAST?
;   IF  compile code: jp
;   ELSE  ['] branch ,BRANCH
;   THEN  ,DEST ; IMMEDIATE
;   unconditional backward branch
    head AGAIN,5,"AGAIN",docolon,1,0
	defw FASTQ
	defw QBRANCH,AGAIN1
	defw LIT,$c3,CCOMMA
	defw BRANCH,AGAIN2
.AGAIN1	defw LIT,BRANCH,COMMABRANCH
.AGAIN2	defw COMMADEST,EXIT

;C WHILE    -- adrs         branch for WHILE loop
;   POSTPONE IF ; IMMEDIATE
    head WHILE,5,"WHILE",docode,1,0
        jp IF

;C REPEAT   adrs1 adrs2 --     resolve WHILE loop
;   SWAP POSTPONE AGAIN POSTPONE THEN ; IMMEDIATE
    head REPEAT,6,"REPEAT",docolon,1,0
        defw SWOP,AGAIN,THEN,EXIT

;Z >L   x --   L: -- x        move to leave stack
;   CELL LP +!  LP @ ! ;      (L stack grows up)
    nohead TOL,docode
	ld l,(iy+usr_lp)
	ld h,(iy+usr_lp+1)
	ld (hl),c
	inc hl
	ld (hl),b
	inc hl
	ld (iy+usr_lp),l
	ld (iy+usr_lp+1),h
	pop bc
	next

;Z L>   -- x   L: x --      move from leave stack
;   LP @ @  CELL NEGATE LP +! ;
    nohead LFROM,docode
	push bc
	ld l,(iy+usr_lp)
	ld h,(iy+usr_lp+1)
	dec hl
	ld b,(hl)
	dec hl
	ld c,(hl)
	ld (iy+usr_lp),l
	ld (iy+usr_lp+1),h
	next

;C DO       -- adrs   L: -- 0
;   ['] xdo ,XT   HERE     target for bwd branch
;   0 >L ; IMMEDIATE           marker for LEAVEs
    head DO,2,"DO",docolon,1,0
        defw LIT,XDO,COMMAXT,HERE
        defw ZERO,TOL,EXIT

;X ?DO      -- addr1  L: -- 0 addr2
;   0 >L			marker for LEAVEs
;   POSTPONE 2DUP POSTPONE =
;   POSTPONE IF POSTPONE 2DROP
;   POSTPONE ELSE >L		store conditional execution on LEAVE stack
;   ['] (do) COMPILE, HERE ; IMMEDIATE
    head QUESTIONDO,3,"?DO",docolon,1,0
	defw ZERO,TOL
	defw LIT,TWODUP,COMMAXT,LIT,EQUAL,COMMAXT
	defw IF,LIT,TWODROP,COMMAXT
	defw ELSE,TOL
	defw LIT,XDO,COMMAXT,HERE,EXIT

;Z ENDLOOP   adrs xt --   L: 0 a1 a2 .. aN --
;   ,BRANCH  ,DEST                backward loop
;   BEGIN L> ?DUP WHILE POSTPONE THEN REPEAT ;
;                                 resolve LEAVEs
; This is a common factor of LOOP and +LOOP.
    nohead ENDLOOP,docolon
        defw COMMABRANCH,COMMADEST
.LOOP1  defw LFROM,QDUP,QBRANCH,LOOP2
        defw THEN,BRANCH,LOOP1
.LOOP2  defw EXIT

;C LOOP    adrs --   L: 0 a1 a2 .. aN --
;   ['] xloop ENDLOOP ;  IMMEDIATE
    head LOOP,4,"LOOP",docolon,1,0
        defw LIT,XLOOP,ENDLOOP,EXIT

;C +LOOP   adrs --   L: 0 a1 a2 .. aN --
;   ['] xplusloop ENDLOOP ;  IMMEDIATE
    head PLUSLOOP,5,"+LOOP",docolon,1,0
        defw LIT,XPLUSLOOP,ENDLOOP,EXIT

;C LEAVE    --    L: -- adrs
;   ['] UNLOOP ,XT
;   ['] branch ,BRANCH   HERE DUP ,DEST  >L
;   ; IMMEDIATE      unconditional forward branch
    head LEAVE,5,"LEAVE",docolon,1,0
        defw LIT,UNLOOP,COMMAXT
        defw LIT,BRANCH,COMMABRANCH
        defw HERE,DUP,COMMADEST,TOL,EXIT


;X CASE   C: -- csys
;   POSTPONE TRUE POSTPONE IF ; IMMEDIATE
    head CASE,4,"CASE",docolon,1,0
	defw LIT,TRUE,COMMAXT
	defw IF,EXIT

;X OF     C: -- ofsys		Runtime: ( x1 x2 -- | x1 )
;   POSTPONE OVER POSTPONE = POSTPONE IF POSTPONE DROP ; IMMEDIATE
    head OF,2,"OF",docolon,1,0
	defw LIT,OVER,COMMAXT
	defw LIT,EQUAL,COMMAXT
	defw IF
	defw LIT,DROP,COMMAXT
	defw EXIT

;X ENDOF  C: csys1 ofsys -- csys2
;   SWAP POSTPONE THEN POSTPONE ELSE ; IMMEDIATE
    head ENDOF,5,"ENDOF",docolon,1,0
        defw SWOP,THEN,ELSE,EXIT

;X ENDCASE  C: csys --		Runtime: ( x -- )
;   POSTPONE DROP POSTPONE THEN ; IMMEDIATE
    head ENDCASE,7,"ENDCASE",docolon,1,0
	defw LIT,DROP,COMMAXT
	defw THEN,EXIT

;Z FOR		C: -- addr	Runtime: ( u --  R: -- u )
;   POSTPONE >R HERE ; IMMEDIATE
    head FOR,3,"FOR",docolon,1,0
	defw LIT,TOR,COMMAXT
	defw HERE,EXIT

;Z (STEP)	( R: uN uN-1..u1 u0 -- )
; This is the inline code version
.brnexts
	ld l,(ix+0)
	ld h,(ix+1)
	dec hl
	ld (ix+0),l
	ld (ix+1),h
	ld a,h
	or l
.bnxtj1	jp nz,0		; this gets patched to the loop address
	inc ix
	inc ix
.brnextn

;Z (STEP)
; This is the standard colon definition version
    nohead BRSTEP,docode
	ld l,(ix+0)
	ld h,(ix+1)
	dec hl
	ld a,h
	or l
	jr z,BRNXT1
	ld (ix+0),l
	ld (ix+1),h
	ex de,hl
	ld e,(hl)
	inc hl
	ld d,(hl)		; DE=inline branch value
	next
.BRNXT1 inc ix			; discard loop value
	inc ix
	inc de			; skip inline branch
	inc de
	next

;Z STEP		C: addr --	Runtime: ( R: uN uN-1..u1 u0 -- )
;   FAST?
;   IF  HERE brnextn-brnexts ALLOT
;       brnexts OVER brnextn-brnexts CMOVE
;       bnxtj1+1-brnexts + !
;   ELSE  ['] brstep ,BRANCH ,DEST
;   THEN ; IMMEDIATE
    head STEP,4,"STEP",docolon,1,0
	defw FASTQ
	defw QBRANCH,STEP1
	defw HERE,LIT,brnextn-brnexts,ALLOT
	defw LIT,brnexts,OVER,LIT,brnextn-brnexts,CMOVE
	defw LIT,bnxtj1+1-brnexts,PLUS,STORE
	defw EXIT
.STEP1	defw LIT,BRSTEP,COMMABRANCH,COMMADEST,EXIT


; OTHER OPERATIONS ==============================

;X WITHIN   n1|u1 n2|u2 n3|u3 -- f   n2<=n1<n3?
;  OVER - >R - R> U< ;          per ANS document
    head WITHIN,6,"WITHIN",docolon,0,0
        defw OVER,MINUS,TOR,MINUS,RFROM,ULESS,EXIT

;C MOVE    addr1 addr2 u --     smart move
;             VERSION FOR 1 ADDRESS UNIT = 1 CHAR
;  >R 2DUP SWAP DUP R@ +     -- ... dst src src+n
;  WITHIN IF  R> CMOVE>        src <= dst < src+n
;       ELSE  R> CMOVE  THEN ;          otherwise
    head MOVE,4,"MOVE",docolon,0,0
        defw TOR,TWODUP,SWOP,DUP,RFETCH,PLUS
        defw WITHIN,QBRANCH,MOVE1
        defw RFROM,CMOVEUP,EXIT
.MOVE1  defw RFROM,CMOVE
.MOVE2  defw EXIT

;C DEPTH    -- +n        number of items on stack
;   SP@ S0 SWAP - 2/ ;   16-BIT VERSION!
    head DEPTH,5,"DEPTH",docolon,0,0
        defw SPFETCH,S0,SWOP,MINUS,TWOSLASH,EXIT

;Z ENVVAR i*x n "name" --  Create new environment variable
;   BL WORD DUP C@ 1+ CHARS ALLOT
;   ALIGN E0 DUP @ , ! DUP ,
;   FOR , STEP ;
    head ENVVAR,6,"ENVVAR",docolon,0,0
	defw BL,WORD,DUP,CFETCH,ONEPLUS,ALLOT
	defw LIT,usr_e0,DUP,FETCH,COMMA,STORE,DUP,COMMA
	defw TOR
.ENVV2  defw COMMA,BRSTEP,ENVV2,EXIT


;C ENVIRONMENT?  c-addr u -- false   system query
;                         -- i*x true
;   E0
;   BEGIN
;     @ DUP >R
;   WHILE
;     2DUP R@ CHAR+ SWAP S=
;     OVER R@ C@ <> OR
;     R> COUNT CHARS + ALIGNED
;     SWAP 0= IF  NIP NIP
;		  CELL+ DUP @ DUP
;		  >R CELLS + R>
;		  FOR  DUP @ SWAP 1 CELLS -  STEP
;		  DROP TRUE EXIT
;	      THEN
;   REPEAT
;   R> DROP 2DROP FALSE ; 
    head ENVIRONMENTQ,12,"ENVIRONMENT?",docolon,0,0
        defw LIT,usr_e0
.ENVQ2  defw FETCH,DUP,TOR
	defw QBRANCH,ENVQ5
	defw TWODUP,RFETCH,CHARPLUS,SWOP,SEQUAL
	defw OVER,RFETCH,CFETCH,NOTEQUAL,OR
	defw RFROM,COUNT,PLUS
	defw SWOP,ZEROEQUAL
	defw QBRANCH,ENVQ4
	defw NIP,NIP,CELLPLUS,DUP,FETCH,DUP
	defw TOR,CELLS,PLUS,RFROM
	defw TOR
.ENVQ3  defw DUP,FETCH,SWOP,ONE,CELLS,MINUS,BRSTEP,ENVQ3
	defw DROP,TRUE,EXIT
.ENVQ4  defw BRANCH,ENVQ2
.ENVQ5  defw RFROM,DROP,TWODROP,FALSE,EXIT


; STRING WORDS ==================================

;X COMPARE	caddr1 u1 caddr2 u2 -- n	ANS string compare
;   ROT 2DUP UMIN >R
;   2SWAP R> S=
;   DUP IF  NIP  ELSE  SWAP ROT - POLARITY  THEN
;   NIP ;
    head COMPARE,7,"COMPARE",docolon,0,0
	defw ROT,TWODUP,UMIN,TOR
	defw TWOSWAP,RFROM,SEQUAL
	defw DUP,QBRANCH,compr1
	defw NIP,BRANCH,compr2
.compr1 defw SWOP,ROT,MINUS,POLARITY
.compr2 defw NIP,EXIT

;X SLITERAL	Compile: caddr u --	Run: -- caddr u
;   POSTPONE (S")
;   DUP C,
;   >R HERE R@ CMOVE R> ALLOT ; IMMEDIATE
    head SLITERAL,8,"SLITERAL",docolon,1,0
	defw LIT,XSQUOTE,COMMAXT
	defw DUP,CCOMMA
	defw TOR,HERE,RFETCH,CMOVE,RFROM,ALLOT,EXIT

;X -TRAILING	caddr u1 -- caddr u2	Remove trailing blanks
;   BEGIN  2DUP 1- CHARS + C@ BL =  OVER 0<> AND
;   WHILE  1-  REPEAT ;
    head MINUSTRAILING,9,"-TRAILING",docolon,0,0
.trail1 defw TWODUP,ONEMINUS,PLUS,CFETCH,BL,EQUAL
	defw OVER,ZERONOTEQUAL,AND
	defw QBRANCH,trail2
	defw ONEMINUS
	defw BRANCH,trail1
.trail2 defw EXIT


; VALUE WORDS ===================================

;X VALUE	x "<spaces>name" --	create value
;   CREATE , DOES> (codefragment)
    head VALUE,5,"VALUE",docolon,0,0
	defw CREATE,COMMA,XDOES
.dovalue
	pop hl		; pointer to data field
	push bc		; stack old TOS
	ld c,(hl)
	inc hl
	ld h,(hl)
	ld l,c		; HL=address of data
	ld c,(hl)
	inc hl
	ld b,(hl)	; TOS=value
	next	

;X TO		x "<spaces>name" --	set a value
;   ' >BODY STATE C@ IF  ,LIT POSTPONE !
;		    ELSE  !  THEN ; IMMEDIATE
    head TO,2,"TO",docolon,1,0
	defw TICK,TOBODY,STATE,CFETCH
	defw QBRANCH,to2
	defw COMMALIT,LIT,STORE,COMMAXT
	defw EXIT
.to2    defw STORE
.to3    defw EXIT

; DEFERRED WORDS ================================



;Z IS		xt "<spaces>name" --	set a deferred word
;   POSTPONE TO ; IMMEDIATE
    head IS,2,"IS",docode,1,0
	jp TO

;C EMIT		c --
;   DEFER EMIT
    head EMIT,4,"EMIT",dodefer,0,0
	defw usr_emit

;C KEY		-- c
;   DEFER KEY
    head KEY,3,"KEY",dodefer,0,0
	defw usr_key

;C TYPE		c-addr u --
;   DEFER TYPE
    head TYPE,4,"TYPE",dodefer,0,0
	defw usr_type

;C (ACC_MAIL)	maxlen curlen caddr pos -- maxlen curlen caddr pos  R: x y ret
;   DEFER (ACC_MAIL)
    head BRACCMAIL,10,"(ACC_MAIL)",dodefer,0,0
	defw usr_braccmail

;C (ACC_EVT)	maxlen curlen caddr pos evt--maxlen curlen caddr pos  R: x y ret
;   DEFER (ACC_EVT)
    head BRACCEVT,9,"(ACC_EVT)",dodefer,0,0
	defw usr_braccevt

;Z (COLD)	--	cold start
;   DEFER (COLD)
    head BRCOLD,6,"(COLD)",dodefer,0,0
	defw usr_cold


; UTILITY WORDS AND STARTUP =====================

;X ERASE	caddr u --
;   0 FILL ;
    head ERASE,5,"ERASE",docolon,0,0
	defw ZERO,FILL,EXIT

;X BLANK	caddr u --
;   BL FILL ;
    head BLANK,5,"BLANK",docolon,0,0
	defw BL,FILL,EXIT

; WORDS is in CAMELLOW.S for the Z88 version. This is the standard definition.
;X WORDS    --          list all words in dict.
;   SORDER @
;   IF  SORDER CELL+ @
;       BEGIN  @ ?DUP
;       WHILE  DUP COUNT TYPE SPACE
;              NFA>LFA
;       REPEAT
;   THEN ;
;    head WORDS,5,"WORDS",docolon,0,0
;	defw SORDER,FETCH,QBRANCH,WDS2
;	defw SORDER,CELLPLUS,FETCH
;WDS1:	defw FETCH,QDUP
;	defw QBRANCH,WDS2
;	defw DUP,COUNT,TYPE,SPACE
;	defw NFATOLFA
;	defw BRANCH,WDS1
;WDS2:	defw EXIT

;X .S      --           print stack contents
;   SP@ S0 - IF
;       SP@ S0 2 - DO I @ U. -2 +LOOP
;   THEN ;
    head DOTS,2,".S",docolon,0,0
        defw SPFETCH,S0,MINUS,QBRANCH,DOTS2
        defw SPFETCH,S0,LIT,2,MINUS,XDO
.DOTS1  defw II,FETCH,UDOT,LIT,-2,XPLUSLOOP,DOTS1
.DOTS2  defw EXIT

;X ?	a-addr --	print cell contents
;   @ . ;
    head QUESTION,1,"?",docolon,0,0
	defw FETCH,DOT,EXIT

;Z COLD     --      cold start Forth system
;   R0 RP! TERMINAL UP!
;   U0 128 ERASE BLOCKID 23 ERASE EMPTY-BUFFERS erase all user vars (2 lots)
;   UINIT U0 #INIT CMOVE			init user area
;   S0 SP! L0 LP ! 1 #TASKS !
;   (COLD) BYE ;		cold start, exit to operating system if returns
    head COLD,4,"COLD",docode,0,0
	ld ix,usr_rstacktop	; RP for terminal task - must do before docolon
	ld iy,pertask_start	; UP for terminal task
	call docolon
	defw U0,LIT,128,ERASE,BLOCKID,LIT,23,ERASE,EMPTYBUFFERS
        defw UINIT,U0,NINIT,CMOVE
	defw S0,SPSTORE,LIT,usr_rstackbot,LP,STORE,ONE,NUMTASKS,STORE
	defw BRCOLD,BYE

